==> nmdc-1.0.0.tgz: package/nmdc.js <==
/* _  ___  _| _    * _
  | )[ | )(_](_ *  |_)       Copyright (c) 2012 -- the nmdc.js author
                 ._|   
Permission to use, copy, modify, and/or distribute this software for any
  purpose with or without fee is hereby granted, provided that the above
  copyright notice and this permission notice appear in all copies.
*/

var net = require('net');

// #################
// Constructor
// #################

exports.Nmdc = function(options, onConnect) {

	// Store handlers
	this.onConnect    = onConnect || function(){};	
	this.onPublic     = function(u,m){};
	this.onPrivate    = function(u,m){};
	this.onUserJoin   = function(u){};
	this.onUserPart   = function(u){};
	this.onUserUpdate = function(u){};
	this.onDebug      = function(s){};
	this.onClosed     = function(){};
	
	this.opts = {
		address: '127.0.0.1',
		port: 411,
		password: '',
		auto_reconnect: false,
		encoding: 'utf8',
		
		nick: 'nmdc.js_user',
		desc: '',
		tag: "nmdc.js 1.0",
		share: 0,		
	};
	
	if (typeof(options) !== 'undefined') {
		for (i in options) this.opts[i] = options[i];
	}
	
	this.users = {};
	this.hubName = '';	
	this.lastError = '';
	
	this.nmdc_connected = false;	
	this.nmdc_partial = '';
	
	/* setTimeout(function() {
		if (this.nmdc_connected) sock.write('|'); // keepalive
	}, 60000); */
	
	return this.reconnect();
}

// #################
// Methods
// #################

exports.Nmdc.prototype.say = function(message) {
	this.sock.write('<'+this.opts.nick+'> '+nmdc_escape(message)+'|');
	return this;
}

exports.Nmdc.prototype.pm = function(user, message) {
	this.sock.write(
		'$To: '+user+' From: '+this.opts.nick+' $<'+this.opts.nick+'> '+
		nmdc_escape(message)+'|'
	);
	return this;
}

exports.Nmdc.prototype.disconnect = function() {
	this.sock.destroy();
	this.onClosed();
	return this;
}

exports.Nmdc.prototype.reconnect = function() {
	var self = this;
	
	if (typeof(this.sock) !== 'undefined') {
		this.disconnect();
	}
	
	this.sock = net.createConnection(this.opts.port, this.opts.address);
	this.sock.setEncoding(this.opts.encoding);
	this.sock.on('connect', function() {
		self.onDebug('NMDC: Socket opened.');
	});
	
	this.sock.on('end', function() {
		self.onDebug('NMDC: Disconnected.');
		self.nmdc_connected = false;
		self.onClosed();
		if (self.opts.auto_reconnect) {
			setTimeout(function(){
				self.reconnect();
			}, 15000);
		}
	});
	
	this.sock.on('data', function(data) {
		self.onDebug('NMDC: '+data);
		var commands = data.split('|');
		
		// Handle protocol buffering
		commands[0] = self.nmdc_partial + commands[0];
		self.nmdc_partial = commands[commands.length - 1];		
		for (var i = 0; i < commands.length - 1; i++) {
			self.nmdc_handle(commands[i]);
		}
		
		//for (i in commands) self.nmdc_handle(commands[i]);		
	});
	
	return this;
}

// #################
// Internal
// #################

exports.Nmdc.prototype.nmdc_handle = function(data) {
	// Short-circuit public chat
	if (data[0] == '<') {
		var rpos = data.indexOf('> ');
		this.onPublic(
			data.substr(1, rpos-1),
			nmdc_unescape(data.substr(rpos+2))
		);
		return this;
	}
	
	// Short-circuit system messages
	if (data[0] != '$') {
		this.onDebug(data);
		return this;
	}

	var cmd = data.split(' ')[0];
	var rem = data.substr(cmd.length + 1);	
	switch (cmd) {
		
		case '$Lock': {
			var key = nmdc_locktokey(rem);
			this.sock.write(
				'$Supports NoGetINFO UserIP2|'+
				'$Key '+key+'|'+
				'$ValidateNick '+this.opts.nick+'|'
			);
		} break;
		
		case '$Hello': {
			if (rem == this.opts.nick) {
				this.sock.write('$Version 1,0091|');
				this.sock.write('$GetNickList|');
				this.sock.write('$MyINFO '+nmdc_getmyinfo(this.opts)+'|');
				
				this.onConnect();
				this.onConnect = function(){}; // don't call again
			} else {
				if (!(rem in this.users)) {
					this.users[rem] = '';
					this.onUserJoin(rem);
				}
			}
		} break;
		
		case '$HubName': {
			this.hubName = rem;
		} break;
		
		case '$ValidateDenide': {
			if (this.opts.password.length) {
				this.lastError = 'Password incorrect.';
			} else {
				this.lastError = 'Nick already in use.';
			}
		} break;
		
		case '$GetPass': {
			this.sock.write('$MyPass '+this.opts.password+'|');
		} break;
		
		case '$Quit': {
			delete this.users[rem];
			this.onUserPart(rem);
		} break;
		
		case '$MyINFO': {
			var user = nmdc_parsemyinfo(rem);
			var nick = user.nick;
			if (!(nick in this.users)) {
				this.users[nick] = '';
				this.onUserJoin(nick);
			}
			this.users[nick] = user;
			this.onUserUpdate(rem);
		} break;
		
		case '$NickList': {
			var users = rem.split('$$');
			for (i in users) {
				var user = users[i];
				if (! user.length) continue;
				if (!(user in this.users)) {
					this.users[user] = '';
					this.onUserJoin(user);
				}
			}
		} break;
		
		case '$To:': {
			var pto = nmdc_parseto(rem);
			this.onPrivate(pto[0], this.nmdc_unescape(pto[1]));
		} break;
		
		// Ignorable:
		case '$Supports':
		case '$UserIP':
		case '$UserCommand':
		case '$UserList':
		case '$OpList':
		{ break; }
		
		default: {
			this.onDebug('NMDC: Unhandled "'+cmd+'"');
		} break;
	}
	
	return this;
}

// #################
// Helpers
// #################


var nmdc_getmyinfo = function(o) {
	return "$ALL "+o.nick+" "+(o.desc.length ? (o.desc+" "):"")+
		"<"+o.tag+",M:P,H:1/0/0,S:5>$ $10  $$"+o.share+"$|";
}

var nmdc_locktokey = function(lock) {
	// Coded by Mardeg
	var nibbleswap = function(bits) {
		return ((bits << 4) & 240) | ((bits >>> 4) & 15);
	}
	var chr = function(b) {
		return (("..0.5.36.96.124.126.").indexOf("."+b+".") > 0) ? 
			"/%DCN"+(0).toPrecision(4-b.toString().length).substr(2)+b+"%/" : 
			String.fromCharCode(b)
		;
	}
	var lock2key = function(lock) {
		var key = chr(nibbleswap(
			lock.charCodeAt(0) ^ lock.charCodeAt(-1) ^ lock.charCodeAt(-2) ^ 5
		));
		for (var i=1; i<lock.length; i++) {
			key += chr(nibbleswap(lock.charCodeAt(i) ^ lock.charCodeAt(i - 1)));
		}
		return key;
	}
	return lock2key(lock);
}

var nmdc_escape = function(str) {
	return (''+str).length ? (''+str).
		replace(/&/g,'&amp;').replace(/\|/g,'&#124;').replace(/\$/g,'&#36;') :
		' ';
}

var nmdc_unescape = function(str) {
	return (''+str).replace(/&#36;/g,'$').
		replace(/&#124;/g,'|').replace(/&amp;/g,'&');
}

var nmdc_parsemyinfo = function(str) {
	// $ALL <nick> <description>$ $<connection><flag>$<e-mail>$<sharesize>$
	var ds = str.indexOf(' ', 6);
	var ret = {
		'nick' : str.substr(5, ds-5),
		'desc' : str.substr(ds+1, str.indexOf('$', 2)-ds),
		'tag'  : '',
		'share': str.substr(str.lastIndexOf('$', str.length-2)).slice(1, -1),
	};
	var tpos = ret.desc.indexOf('<');
	if (tpos !== -1) {
		ret.tag  = ret.desc.substr(tpos+1).slice(0,-2);
		ret.desc = ret.desc.substr(0, tpos-1);
	}
	return ret;
}

var nmdc_parseto = function(str) {
	// recipient From: sender $<sender> message|
	var lpos = str.indexOf('$<');
	var rpos = str.indexOf('> ');
	return [ str.slice(lpos+2, rpos), str.slice(rpos+2) ];
}

==> nmdc-1.1.0a.tgz: package/nmdc.js <==
/* _  ___  _| _    * _
  | )[ | )(_](_ *  |_)       Copyright (c) 2012-2013 -- the nmdc.js author
                 ._|   
Permission to use, copy, modify, and/or distribute this software for any
  purpose with or without fee is hereby granted, provided that the above
  copyright notice and this permission notice appear in all copies.
*/

"use strict";
var net = require('net');

var NMDC_JS_RECONNECT_TIMEOUT = 60*1000;
var NMDC_JS_KEEPALIVE_TIMEOUT = 30*1000;

// #################
// Constructor
// #################

function Nmdc(options, onConnect) {

	// Simulate calling constructor with 'new' if it was omitted
	if (!(this instanceof Nmdc)) {
		return new Nmdc(options, onConnect);
	}

	// Store handlers
	this.onConnect    = onConnect || function(){};
	this.onSystem     = function(s){};
	this.onPublic     = function(u,m){};
	this.onPrivate    = function(u,m){};
	this.onUserJoin   = function(u){};
	this.onUserPart   = function(u){};
	this.onUserUpdate = function(u){};
	this.onDebug      = function(s){};
	this.onClosed     = function(){};
	
	this.opts = {
		address: '127.0.0.1',
		port: 411,
		password: '',
		auto_reconnect: false,
		encoding: 'utf8',
		
		nick: 'nmdc.js_user',
		desc: '',
		tag: "nmdc.js 1.1",
		share: 0,		
	};
	
	if (typeof(options) !== 'undefined') {
		for (var i in options) this.opts[i] = options[i];
	}
	
	this.users = {};
	this.hubName = '';	
	this.lastError = '';
	
	this.nmdc_connected = false;	
	this.nmdc_partial = '';
	
	this._keepAlive = false;
	this._reconnector = false;
	
	return this.reconnect();
}

// #################
// Methods
// #################

Nmdc.prototype.say = function(message, cb) {
	this.sock.write('<'+this.opts.nick+'> '+nmdc_escape(message)+'|', this.opts.encoding, cb);
	return this;
}

Nmdc.prototype.pm = function(user, message, cb) {
	this.sock.write(
		'$To: '+user+' From: '+this.opts.nick+' $<'+this.opts.nick+'> '+
		nmdc_escape(message)+'|',
		this.opts.encoding,
		cb
	);
	return this;
}

Nmdc.prototype.disconnect = function() {
	this.sock.destroy();
	this.onClosed();
	
	clearInterval(this._keepAlive);
	this._keepAlive = false;
	
	clearInterval(this._reconnector);
	this._reconnector = false;
	
	return this;
}

Nmdc.prototype.reconnect = function() {
	var self = this;
	
	if (typeof(this.sock) !== 'undefined') {
		this.disconnect();
	}
	
	this.sock = net.createConnection(this.opts.port, this.opts.address);
	this.sock.setEncoding(this.opts.encoding);
	this.sock.on('connect', function() {
		self.onDebug('NMDC: Socket opened.');
	});
	
	this.sock.on('end', function() {
		self.onDebug('NMDC: Disconnected.');
		self.nmdc_connected = false; // trigger auto reconnect
		self.sock.destroy();
		self.onClosed();
	});
	
	this.sock.on('error', function() {
		self.onDebug('Connection error');
		if (self.nmdc_connected) {
			self.nmdc_connected = false; // trigger auto reconnect
			self.sock.destroy();
			self.onClosed();
		}
	});
	
	this.sock.on('data', function(data) {
		self.onDebug('NMDC: '+data);
		var commands = data.split('|');
		
		// Handle protocol buffering
		commands[0] = self.nmdc_partial + commands[0];
		self.nmdc_partial = commands[commands.length - 1];		
		for (var i = 0; i < commands.length - 1; i++) {
			self.nmdc_handle(commands[i]);
		}
			
	});
	
	// Handle keepalives
	this._keepAlive = setInterval(function() {
		if (this.nmdc_connected) {
			self.onDebug('keepalive');
			sock.write('|');
		}
	}, NMDC_JS_KEEPALIVE_TIMEOUT);
	
	// Handle auto reconnect
	if (this.opts.auto_reconnect) {
		this.__reconnector = setInterval(function() {
			if (! self.nmdc_connected) {
				self.onDebug('Reconnecting...');
				self.reconnect();
			}
		}, NMDC_JS_RECONNECT_TIMEOUT);
	}
	
	return this;
}

// #################
// Internal
// #################

Nmdc.prototype.nmdc_handle = function(data) {
	
	// Short-circuit public chat
	if (data[0] == '<') {
		var rpos = data.indexOf('> ');
		this.onPublic(
			data.substr(1, rpos-1),
			nmdc_unescape(data.substr(rpos+2))
		);
		return this;
	}
	
	if (data[0] == '*') {
		this.onSystem(data);
		return this;
	}
	
	// Short-circuit system messages
	if (data[0] != '$') {
		this.onDebug(data);
		return this;
	}
	
	var cmd = data.split(' ')[0];
	var rem = data.substr(cmd.length + 1);	
	switch (cmd) {
		
		case '$Lock': {
			var key = nmdc_locktokey(rem);
			this.sock.write(
				'$Supports NoGetINFO UserIP2|'+
				'$Key '+key+'|'+
				'$ValidateNick '+this.opts.nick+'|'
			);
		} break;
		
		case '$Hello': {
			if (rem == this.opts.nick) {
				
				// Handshake
				this.sock.write('$Version 1,0091|');
				this.sock.write('$GetNickList|');
				this.sock.write('$MyINFO '+nmdc_getmyinfo(this.opts)+'|');
				
			} else {
				if (!(rem in this.users)) {
					this.users[rem] = '';
					this.onUserJoin(rem);
				}
			}
		} break;
		
		case '$HubName': {
			this.hubName = rem;
		} break;
		
		case '$ValidateDenide': {
			if (this.opts.password.length) {
				this.lastError = 'Password incorrect.';
			} else {
				this.lastError = 'Nick already in use.';
			}
		} break;
		
		case '$GetPass': {
			this.sock.write('$MyPass '+this.opts.password+'|');
		} break;
		
		case '$Quit': {
			delete this.users[rem];
			this.onUserPart(rem);
		} break;
		
		case '$MyINFO': {
			var user = nmdc_parsemyinfo(rem);
			var nick = user.nick;
			if (!(nick in this.users)) {
				this.users[nick] = '';
				this.onUserJoin(nick);
			}
			this.users[nick] = user;
			this.onUserUpdate(rem);
		} break;
		
		case '$NickList': {
			var users = rem.split('$$');
			for (var i in users) {
				var user = users[i];
				if (! user.length) continue;
				if (!(user in this.users)) {
					this.users[user] = '';
					this.onUserJoin(user);
				}
			}
		} break;
		
		case '$To:': {
			var pto = nmdc_parseto(rem);
			this.onPrivate(pto[0], nmdc_unescape(pto[1]));
		} break;
		
		case '$UserIP': {
			// Final message in PtokaX connection handshake - trigger connection
			//  callback. This might not always be the case for other hubsofts?
					
			if (! this.nmdc_connected) {
				this.onConnect(); // Only call once per connection
			}
			this.nmdc_connected = true;			
		}
		
		// Ignorable:
		case '$Supports':
		case '$UserCommand':
		case '$UserList':
		case '$OpList':
		{ break; }
		
		default: {
			this.onDebug('NMDC: Unhandled "'+cmd+'"');
		} break;
	}
	
	return this;
}

// #################
// Exports
// #################

exports.Nmdc = Nmdc;

// #################
// Helpers
// #################


var nmdc_getmyinfo = function(o) {
	return "$ALL "+o.nick+" "+(o.desc.length ? (o.desc+" "):"")+
		"<"+o.tag+",M:P,H:1/0/0,S:5>$ $10  $$"+o.share+"$|";
}

var nmdc_locktokey = function(lock) {
	// Coded by Mardeg
	var nibbleswap = function(bits) {
		return ((bits << 4) & 240) | ((bits >>> 4) & 15);
	}
	var chr = function(b) {
		return (("..0.5.36.96.124.126.").indexOf("."+b+".") > 0) ? 
			"/%DCN"+(0).toPrecision(4-b.toString().length).substr(2)+b+"%/" : 
			String.fromCharCode(b)
		;
	}
	var lock2key = function(lock) {
		var key = chr(nibbleswap(
			lock.charCodeAt(0) ^ lock.charCodeAt(-1) ^ lock.charCodeAt(-2) ^ 5
		));
		for (var i=1; i<lock.length; i++) {
			key += chr(nibbleswap(lock.charCodeAt(i) ^ lock.charCodeAt(i - 1)));
		}
		return key;
	}
	return lock2key(lock);
}

var nmdc_escape = function(str) {
	return (''+str).length ? (''+str).
		replace(/&/g,'&amp;').replace(/\|/g,'&#124;').replace(/\$/g,'&#36;') :
		' ';
}

var nmdc_unescape = function(str) {
	return (''+str).replace(/&#36;/g,'$').
		replace(/&#124;/g,'|').replace(/&amp;/g,'&');
}

var nmdc_parsemyinfo = function(str) {
	// $ALL <nick> <description>$ $<connection><flag>$<e-mail>$<sharesize>$
	var ds = str.indexOf(' ', 6);
	var ret = {
		'nick' : str.substr(5, ds-5),
		'desc' : str.substr(ds+1, str.indexOf('$', 2)-ds),
		'tag'  : '',
		'share': str.substr(str.lastIndexOf('$', str.length-2)).slice(1, -1),
	};
	var tpos = ret.desc.indexOf('<');
	if (tpos !== -1) {
		ret.tag  = ret.desc.substr(tpos+1).slice(0,-2);
		ret.desc = ret.desc.substr(0, tpos-1);
	}
	return ret;
}

var nmdc_parseto = function(str) {
	// recipient From: sender $<sender> message|
	var lpos = str.indexOf('$<');
	var rpos = str.indexOf('> ');
	return [ str.slice(lpos+2, rpos), str.slice(rpos+2) ];
}
